@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 """
 irc.server
 
@@ -47,9 +45,8 @@ will ever be connected to by the public.
 import argparse
 import errno
 import logging
-import socket
-import select
 import re
+import select
 import socketserver
 import typing
 
@@ -57,6 +54,7 @@ import jaraco.logging
 from jaraco.stream import buffer
 
 import irc.client
+
 from . import events
 
 log = logging.getLogger(__name__)
@@ -147,8 +145,8 @@ class IRCClient(socketserver.BaseRequestHandler):
     def _handle_incoming(self):
         try:
             data = self.request.recv(1024)
-        except Exception:
-            raise self.Disconnect()
+        except Exception as err:
+            raise self.Disconnect() from err
 
         if not data:
             raise self.Disconnect()
@@ -159,25 +157,26 @@ class IRCClient(socketserver.BaseRequestHandler):
             self._handle_line(line)
 
     def _handle_line(self, line):
+        response = None
+
         try:
-            log.debug('from %s: %s' % (self.client_ident(), line))
+            log.debug(f'from {self.client_ident()}: {line}')
             command, sep, params = line.partition(' ')
-            handler = getattr(self, 'handle_%s' % command.lower(), None)
+            handler = getattr(self, f'handle_{command.lower()}', None)
             if not handler:
-                _tmpl = 'No handler for command: %s. Full line: %s'
-                log.info(_tmpl % (command, line))
+                log.info(f'No handler for command: {command}. Full line: {line}')
                 raise IRCError.from_name(
-                    'unknowncommand', '%s :Unknown command' % command
+                    'unknowncommand', f'{command} :Unknown command'
                 )
             response = handler(params)
         except AttributeError as e:
             log.error(str(e))
             raise
         except IRCError as e:
-            response = ':%s %s %s' % (self.server.servername, e.code, e.value)
-            log.error(response)
+            response = f':{self.server.servername} {e.code} {e.value}'
+            log.warning(response)
         except Exception as e:
-            response = ':%s ERROR %r' % (self.server.servername, e)
+            response = f':{self.server.servername} ERROR {e!r}'
             log.error(response)
             raise
 
@@ -188,9 +187,9 @@ class IRCClient(socketserver.BaseRequestHandler):
         log.debug('to %s: %s', self.client_ident(), msg)
         try:
             self.request.send(msg.encode('utf-8') + b'\r\n')
-        except socket.error as e:
+        except OSError as e:
             if e.errno == errno.EPIPE:
-                raise self.Disconnect()
+                raise self.Disconnect() from e
             else:
                 raise
 
@@ -202,7 +201,7 @@ class IRCClient(socketserver.BaseRequestHandler):
 
         # Valid nickname?
         if re.search(r'[^a-zA-Z0-9\-\[\]\'`^{}_]', nick):
-            raise IRCError.from_name('erroneusnickname', ':%s' % nick)
+            raise IRCError.from_name('erroneusnickname', f':{nick}')
 
         if self.server.clients.get(nick, None) == self:
             # Already registered to user
@@ -210,7 +209,7 @@ class IRCClient(socketserver.BaseRequestHandler):
 
         if nick in self.server.clients:
             # Someone else is using the nick
-            raise IRCError.from_name('nicknameinuse', 'NICK :%s' % (nick))
+            raise IRCError.from_name('nicknameinuse', f'NICK :{nick}')
 
         if not self.nick:
             # New connection and nick is available; register and send welcome
@@ -218,22 +217,21 @@ class IRCClient(socketserver.BaseRequestHandler):
             self.nick = nick
             self.server.clients[nick] = self
             msg = f"Welcome to {__name__} v{irc._get_version()}."
-            response = ':%s %s %s :%s' % (
+            response = ':{} {} {} :{}'.format(
                 self.server.servername,
                 events.codes['welcome'],
                 self.nick,
                 msg,
             )
             self.send_queue.append(response)
-            response = ':%s 376 %s :End of MOTD command.' % (
-                self.server.servername,
-                self.nick,
+            response = (
+                f':{self.server.servername} 376 {self.nick} :End of MOTD command.'
             )
             self.send_queue.append(response)
             return
 
         # Nick is available. Change the nick.
-        message = ':%s NICK :%s' % (self.client_ident(), nick)
+        message = f':{self.client_ident()} NICK :{nick}'
 
         self.server.clients.pop(self.nick)
         self.nick = nick
@@ -281,7 +279,7 @@ class IRCClient(socketserver.BaseRequestHandler):
             # Valid channel name?
             if not re.match('^#([a-zA-Z0-9_])+$', r_channel_name):
                 raise IRCError.from_name(
-                    'nosuchchannel', '%s :No such channel' % r_channel_name
+                    'nosuchchannel', f'{r_channel_name} :No such channel'
                 )
 
             # Add user to the channel (create new channel if not exists)
@@ -294,47 +292,53 @@ class IRCClient(socketserver.BaseRequestHandler):
             self.channels[channel.name] = channel
 
             # Send the topic
-            response_join = ':%s TOPIC %s :%s' % (
-                channel.topic_by,
-                channel.name,
-                channel.topic,
-            )
+            response_join = f':{channel.topic_by} TOPIC {channel.name} :{channel.topic}'
             self.send_queue.append(response_join)
 
             # Send join message to everybody in the channel, including yourself
             # and send user list of the channel back to the user.
-            response_join = ':%s JOIN :%s' % (self.client_ident(), r_channel_name)
+            response_join = f':{self.client_ident()} JOIN :{r_channel_name}'
             for client in channel.clients:
                 client.send_queue.append(response_join)
 
             nicks = [client.nick for client in channel.clients]
-            _vals = (self.server.servername, self.nick, channel.name, ' '.join(nicks))
-            response_userlist = ':%s 353 %s = %s :%s' % _vals
+            response_userlist = f':{self.server.servername} 353 {self.nick} = {channel.name} :{" ".join(nicks)}'
             self.send_queue.append(response_userlist)
 
-            _vals = self.server.servername, self.nick, channel.name
-            response = ':%s 366 %s %s :End of /NAMES list' % _vals
+            response = f':{self.server.servername} 366 {self.nick} {channel.name} :End of /NAMES list'
             self.send_queue.append(response)
 
     def handle_privmsg(self, params):
         """
         Handle sending a private message to a user or channel.
         """
+        self._send_msg('PRIVMSG', params)
+
+    def handle_notice(self, params):
+        """
+        Handle sending a notice to a user or channel.
+        """
+        self._send_msg('NOTICE', params)
+
+    def _send_msg(self, cmd, params):
+        """
+        A generic message handler (e.g. PRIVMSG and NOTICE)
+        """
         target, sep, msg = params.partition(' ')
         if not msg:
-            raise IRCError.from_name('needmoreparams', 'PRIVMSG :Not enough parameters')
+            raise IRCError.from_name('needmoreparams', cmd + ' :Not enough parameters')
 
-        message = ':%s PRIVMSG %s %s' % (self.client_ident(), target, msg)
+        message = f':{self.client_ident()} {cmd} {target} {msg}'
         if target.startswith('#') or target.startswith('$'):
             # Message to channel. Check if the channel exists.
             channel = self.server.channels.get(target)
             if not channel:
-                raise IRCError.from_name('nosuchnick', 'PRIVMSG :%s' % target)
+                raise IRCError.from_name('nosuchnick', cmd + f' :{target}')
 
             if channel.name not in self.channels:
                 # The user isn't in the channel.
                 raise IRCError.from_name(
-                    'cannotsendtochan', '%s :Cannot send to channel' % channel.name
+                    'cannotsendtochan', f'{channel.name} :Cannot send to channel'
                 )
 
             self._send_to_others(message, channel)
@@ -342,7 +346,7 @@ class IRCClient(socketserver.BaseRequestHandler):
             # Message to user
             client = self.server.clients.get(target, None)
             if not client:
-                raise IRCError.from_name('nosuchnick', 'PRIVMSG :%s' % target)
+                raise IRCError.from_name('nosuchnick', cmd + f' :{target}')
 
             client.send_queue.append(message)
 
@@ -363,21 +367,17 @@ class IRCClient(socketserver.BaseRequestHandler):
 
         channel = self.server.channels.get(channel_name)
         if not channel:
-            raise IRCError.from_name('nosuchnick', 'PRIVMSG :%s' % channel_name)
+            raise IRCError.from_name('nosuchnick', f'PRIVMSG :{channel_name}')
         if channel.name not in self.channels:
             # The user isn't in the channel.
             raise IRCError.from_name(
-                'cannotsendtochan', '%s :Cannot send to channel' % channel.name
+                'cannotsendtochan', f'{channel.name} :Cannot send to channel'
             )
 
         if topic:
             channel.topic = topic.lstrip(':')
             channel.topic_by = self.nick
-        message = ':%s TOPIC %s :%s' % (
-            self.client_ident(),
-            channel_name,
-            channel.topic,
-        )
+        message = f':{self.client_ident()} TOPIC {channel_name} :{channel.topic}'
         return message
 
     def handle_part(self, params):
@@ -389,22 +389,21 @@ class IRCClient(socketserver.BaseRequestHandler):
                 # Send message to all clients in all channels user is in, and
                 # remove the user from the channels.
                 channel = self.server.channels.get(pchannel.strip())
-                response = ':%s PART :%s' % (self.client_ident(), pchannel)
+                response = f':{self.client_ident()} PART :{pchannel}'
                 if channel:
                     for client in channel.clients:
                         client.send_queue.append(response)
                 channel.clients.remove(self)
                 self.channels.pop(pchannel)
             else:
-                _vars = self.server.servername, pchannel, pchannel
-                response = ':%s 403 %s :%s' % _vars
+                response = f':{self.server.servername} 403 {pchannel} :{pchannel}'
                 self.send_queue.append(response)
 
     def handle_quit(self, params):
         """
         Handle the client breaking off the connection with a QUIT command.
         """
-        response = ':%s QUIT :%s' % (self.client_ident(), params.lstrip(':'))
+        response = ':{} QUIT :{}'.format(self.client_ident(), params.lstrip(':'))
         # Send quit message to all clients in all channels user is in, and
         # remove the user from the channels.
         for channel in self.channels.values():
@@ -428,12 +427,9 @@ class IRCClient(socketserver.BaseRequestHandler):
                 print("     ", client.nick, client)
 
     def handle_ison(self, params):
-        response = ':%s 303 %s :' % (self.server.servername, self.client_ident().nick)
+        response = f':{self.server.servername} 303 {self.client_ident().nick} :'
         if len(params) == 0 or params.isspace():
-            response = ':%s 461 %s ISON :Not enough parameters' % (
-                self.server.servername,
-                self.client_ident().nick,
-            )
+            response = f':{self.server.servername} 461 {self.client_ident().nick} ISON :Not enough parameters'
             return response
         nickOnline = []
         for nick in params.split(" "):
@@ -457,7 +453,7 @@ class IRCClient(socketserver.BaseRequestHandler):
         the client didn't properly close the connection with PART and QUIT.
         """
         log.info('Client disconnected: %s', self.client_ident())
-        response = ':%s QUIT :EOF from client' % self.client_ident()
+        response = f':{self.client_ident()} QUIT :EOF from client'
         for channel in self.channels.values():
             if self in channel.clients:
                 # Client is gone without properly QUITing or PARTing this
@@ -473,13 +469,7 @@ class IRCClient(socketserver.BaseRequestHandler):
         """
         Return a user-readable description of the client
         """
-        return '<%s %s!%s@%s (%s)>' % (
-            self.__class__.__name__,
-            self.nick,
-            self.user,
-            self.host[0],
-            self.realname,
-        )
+        return f'<{self.__class__.__name__} {self.nick}!{self.user}@{self.host[0]} ({self.realname})>'
 
 
 class IRCServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
@@ -535,9 +525,9 @@ def main():
         _tmpl = 'Listening on {listen_address}:{listen_port}'
         log.info(_tmpl.format(**vars(options)))
         ircserver.serve_forever()
-    except socket.error as e:
+    except OSError as e:
         log.error(repr(e))
-        raise SystemExit(-2)
+        raise SystemExit(-2) from None
 
 
 if __name__ == "__main__":
